from __future__ import annotations

import json
import logging
import re
from collections.abc import Iterator
from typing import Any, cast

import pytest

from acodex._internal.exec import CodexExecArgs
from acodex._internal.thread_core import (
    build_turn_or_raise,
    initial_turn_state,
    normalize_input,
    parse_thread_event_jsonl,
    parse_thread_item,
    reduce_turn_state,
)
from acodex.exceptions import CodexThreadRunError
from acodex.exec import CodexExec
from acodex.thread import Thread
from acodex.types.events import ItemCompletedEvent, ThreadEvent, TurnCompletedEvent, TurnFailedEvent
from acodex.types.input import UserInputLocalImage, UserInputText
from acodex.types.items import (
    AgentMessageItem,
    CommandExecutionItem,
    ErrorItem,
    FileChangeItem,
    McpToolCallItem,
    ReasoningItem,
    TodoListItem,
    WebSearchItem,
)


def test_unknown_event_type_is_skipped_and_logs_warning(caplog: pytest.LogCaptureFixture) -> None:
    thread = _build_fake_thread(
        lines=[
            json.dumps({"type": "thread.started", "thread_id": "thread-1"}),
            json.dumps({"type": "future.event", "payload": {"x": 1}}),
            json.dumps(
                {
                    "type": "turn.completed",
                    "usage": {
                        "input_tokens": 1,
                        "cached_input_tokens": 0,
                        "output_tokens": 1,
                    },
                },
            ),
        ],
    )
    caplog.set_level(logging.WARNING, logger="acodex._internal.thread_core")

    events = list(thread.run_streamed("hello").events)

    assert [event.type for event in events] == ["thread.started", "turn.completed"]
    assert "Unknown thread event type future.event; skipping." in caplog.text


def test_unknown_item_type_event_is_skipped_and_logs_warning(
    caplog: pytest.LogCaptureFixture,
) -> None:
    thread = _build_fake_thread(
        lines=[
            json.dumps({"type": "thread.started", "thread_id": "thread-1"}),
            json.dumps(
                {
                    "type": "item.completed",
                    "item": {"type": "future.item", "id": "future-item"},
                },
            ),
            json.dumps(
                {
                    "type": "turn.completed",
                    "usage": {
                        "input_tokens": 2,
                        "cached_input_tokens": 0,
                        "output_tokens": 3,
                    },
                },
            ),
        ],
    )
    caplog.set_level(logging.WARNING, logger="acodex._internal.thread_core")

    events = list(thread.run_streamed("hello").events)

    assert [event.type for event in events] == ["thread.started", "turn.completed"]
    assert "Unknown thread item type future.item; skipping." in caplog.text
    assert "Unknown thread item for item.completed; skipping event." in caplog.text


def test_parse_thread_event_jsonl_invalid_json_raises_codex_thread_run_error() -> None:
    with pytest.raises(CodexThreadRunError, match="Failed to parse item: not-json"):
        parse_thread_event_jsonl("not-json")


def test_parse_thread_event_jsonl_rejects_non_object_payload() -> None:
    with pytest.raises(CodexThreadRunError, match=re.escape("Failed to parse item: []")):
        parse_thread_event_jsonl("[]")


def test_parse_thread_event_jsonl_wraps_type_errors_from_parsers() -> None:
    line = json.dumps(
        {
            "type": "turn.completed",
            "usage": {
                "input_tokens": "x",
                "cached_input_tokens": 0,
                "output_tokens": 1,
            },
        },
    )

    with pytest.raises(CodexThreadRunError, match=re.escape(f"Failed to parse item: {line}")):
        parse_thread_event_jsonl(line)


def test_normalize_input_handles_string_and_list() -> None:
    from_string = normalize_input("hello")
    assert from_string.prompt == "hello"
    assert from_string.images == []

    from_list = normalize_input(
        [
            UserInputText(text="alpha"),
            UserInputLocalImage(path="/images/a.png"),
            UserInputText(text="beta"),
        ],
    )
    assert from_list.prompt == "alpha\n\nbeta"
    assert from_list.images == ["/images/a.png"]


def test_normalize_input_rejects_invalid_shapes() -> None:
    with pytest.raises(TypeError, match=r"input must be str or list\[UserInput\], got dict"):
        normalize_input(cast("Any", {"type": "text"}))

    with pytest.raises(
        TypeError,
        match="input\\[0\\] must be UserInputText or UserInputLocalImage",
    ):
        normalize_input(cast("Any", [{"type": "text", "text": 1}]))

    with pytest.raises(
        TypeError,
        match="input\\[0\\] must be UserInputText or UserInputLocalImage",
    ):
        normalize_input(cast("Any", [{"type": "future", "x": "y"}]))


def test_parse_thread_item_supports_all_public_item_variants() -> None:
    agent_item = parse_thread_item({"type": "agent_message", "id": "a", "text": "hello"})
    reasoning_item = parse_thread_item({"type": "reasoning", "id": "r", "text": "why"})
    command_item = parse_thread_item(
        {
            "type": "command_execution",
            "id": "c",
            "command": "ls",
            "aggregated_output": "ok",
            "status": "completed",
            "exit_code": 0,
        },
    )
    file_change_item = parse_thread_item(
        {
            "type": "file_change",
            "id": "f",
            "status": "completed",
            "changes": [{"path": "x.py", "kind": "update"}],
        },
    )
    mcp_item = parse_thread_item(
        {
            "type": "mcp_tool_call",
            "id": "m",
            "server": "srv",
            "tool": "tool",
            "arguments": {"x": 1},
            "status": "completed",
            "result": {"content": [{"type": "text", "text": "ok"}], "structured_content": {"a": 1}},
            "error": {"message": "not-used"},
        },
    )
    web_item = parse_thread_item({"type": "web_search", "id": "w", "query": "q"})
    todo_item = parse_thread_item(
        {
            "type": "todo_list",
            "id": "t",
            "items": [{"text": "step", "completed": False}],
        },
    )
    error_item = parse_thread_item({"type": "error", "id": "e", "message": "oops"})

    assert isinstance(agent_item, AgentMessageItem)
    assert isinstance(reasoning_item, ReasoningItem)
    assert isinstance(command_item, CommandExecutionItem)
    assert isinstance(file_change_item, FileChangeItem)
    assert isinstance(mcp_item, McpToolCallItem)
    assert isinstance(web_item, WebSearchItem)
    assert isinstance(todo_item, TodoListItem)
    assert isinstance(error_item, ErrorItem)


def test_parse_item_started_event_is_supported() -> None:
    event = parse_thread_event_jsonl(
        json.dumps(
            {
                "type": "item.started",
                "item": {"type": "agent_message", "id": "a", "text": "hello"},
            },
        ),
    )

    assert event is not None
    assert event.type == "item.started"


def test_parse_mcp_tool_call_allows_missing_result_and_error() -> None:
    item = parse_thread_item(
        {
            "type": "mcp_tool_call",
            "id": "mcp-1",
            "server": "srv",
            "tool": "run",
            "arguments": {"x": 1},
            "status": "in_progress",
        },
    )

    assert isinstance(item, McpToolCallItem)
    assert item.result is None
    assert item.error is None


def test_parse_thread_item_command_execution_exit_code_is_optional() -> None:
    item = parse_thread_item(
        {
            "type": "command_execution",
            "id": "c-1",
            "command": "ls",
            "aggregated_output": "ok",
            "status": "in_progress",
        },
    )

    assert isinstance(item, CommandExecutionItem)
    assert item.exit_code is None


def test_parse_thread_item_rejects_invalid_shapes_with_field_messages() -> None:
    with pytest.raises(TypeError, match="thread item must be an object"):
        parse_thread_item("not-an-object")

    with pytest.raises(TypeError, match=re.escape("web_search.id must be a string")):
        parse_thread_item({"type": "web_search", "id": 123, "query": "q"})

    with pytest.raises(TypeError, match=re.escape("todo_list.items must be an array")):
        parse_thread_item({"type": "todo_list", "id": "t", "items": "nope"})

    with pytest.raises(
        TypeError,
        match=re.escape("todo_list.items[0].completed must be a boolean"),
    ):
        parse_thread_item(
            {
                "type": "todo_list",
                "id": "t",
                "items": [{"text": "task", "completed": "no"}],
            },
        )


def test_parse_thread_event_jsonl_parses_event_variants() -> None:
    started = parse_thread_event_jsonl(json.dumps({"type": "turn.started"}))
    updated = parse_thread_event_jsonl(
        json.dumps(
            {
                "type": "item.updated",
                "item": {"type": "agent_message", "id": "a", "text": "hello"},
            },
        ),
    )
    error = parse_thread_event_jsonl(json.dumps({"type": "error", "message": "fatal"}))

    assert started is not None
    assert started.type == "turn.started"
    assert updated is not None
    assert updated.type == "item.updated"
    assert error is not None
    assert error.type == "error"


def test_turn_state_reduction_builds_turn_and_raises_on_failure() -> None:
    completed_state = initial_turn_state()
    completed_events: list[ThreadEvent] = [
        cast(
            "ItemCompletedEvent",
            parse_thread_event_jsonl(
                json.dumps(
                    {
                        "type": "item.completed",
                        "item": {"type": "agent_message", "id": "a", "text": "hello"},
                    },
                ),
            ),
        ),
        cast(
            "TurnCompletedEvent",
            parse_thread_event_jsonl(
                json.dumps(
                    {
                        "type": "turn.completed",
                        "usage": {
                            "input_tokens": 3,
                            "cached_input_tokens": 1,
                            "output_tokens": 2,
                        },
                    },
                ),
            ),
        ),
    ]
    for event in completed_events:
        completed_state = reduce_turn_state(completed_state, event)

    turn = build_turn_or_raise(completed_state)
    assert turn.final_response == "hello"
    assert turn.usage is not None
    assert turn.usage.output_tokens == 2
    assert len(turn.items) == 1

    failed_state = initial_turn_state()
    failed_event = cast(
        "TurnFailedEvent",
        parse_thread_event_jsonl(
            json.dumps({"type": "turn.failed", "error": {"message": "failed"}}),
        ),
    )
    failed_state = reduce_turn_state(failed_state, failed_event)

    with pytest.raises(CodexThreadRunError, match="failed"):
        build_turn_or_raise(failed_state)


def test_reduce_turn_state_does_not_set_final_response_for_non_agent_item() -> None:
    state = initial_turn_state()
    event = ItemCompletedEvent(item=ReasoningItem(id="r-1", text="thinking"))

    reduced = reduce_turn_state(state, event)

    assert not reduced.final_response


class _FakeExec:
    def __init__(self, *, lines: list[str]) -> None:
        self._lines = lines

    def run(self, args: CodexExecArgs) -> Iterator[str]:
        _ = args
        yield from self._lines


def _build_fake_thread(*, lines: list[str]) -> Thread:
    fake_exec = cast("CodexExec", _FakeExec(lines=lines))
    return Thread(exec=fake_exec, options={}, thread_options={})



from __future__ import annotations

import json
import logging
from collections.abc import Callable
from typing import NamedTuple, cast

from acodex.exceptions import CodexThreadRunError
from acodex.types.events import (
    ItemCompletedEvent,
    ItemStartedEvent,
    ItemUpdatedEvent,
    ThreadError,
    ThreadErrorEvent,
    ThreadEvent,
    ThreadStartedEvent,
    TurnCompletedEvent,
    TurnFailedEvent,
    TurnStartedEvent,
    Usage,
)
from acodex.types.input import Input, UserInputLocalImage, UserInputText
from acodex.types.items import (
    AgentMessageItem,
    CommandExecutionItem,
    CommandExecutionStatus,
    ErrorItem,
    FileChangeItem,
    FileUpdateChange,
    McpToolCallError,
    McpToolCallItem,
    McpToolCallResult,
    McpToolCallStatus,
    PatchApplyStatus,
    PatchChangeKind,
    ReasoningItem,
    ThreadItem,
    TodoItem,
    TodoListItem,
    WebSearchItem,
)
from acodex.types.turn import Turn

logger = logging.getLogger(__name__)


class NormalizedInput(NamedTuple):
    prompt: str
    images: list[str]


def normalize_input(input: Input) -> NormalizedInput:  # noqa: A002
    if isinstance(input, str):
        return NormalizedInput(prompt=input, images=[])

    if not isinstance(input, list):
        raise TypeError(f"input must be str or list[UserInput], got {type(input).__name__}")

    prompt_parts: list[str] = []
    images: list[str] = []

    for index, raw_item in enumerate(input):
        if isinstance(raw_item, UserInputText):
            prompt_parts.append(raw_item.text)
            continue

        if isinstance(raw_item, UserInputLocalImage):
            images.append(raw_item.path)
            continue

        raise TypeError(
            f"input[{index}] must be UserInputText or UserInputLocalImage, got {type(raw_item).__name__}",
        )

    return NormalizedInput(prompt="\n\n".join(prompt_parts), images=images)


def parse_thread_event_jsonl(line: str) -> ThreadEvent | None:
    try:
        payload_obj = json.loads(line)
    except json.JSONDecodeError as error:
        raise CodexThreadRunError(f"Failed to parse item: {line}") from error

    if not isinstance(payload_obj, dict):
        raise CodexThreadRunError(f"Failed to parse item: {line}")

    payload = cast("dict[str, object]", payload_obj)
    try:
        event_type = _require_str(payload.get("type"), "event.type")
        parser = _EVENT_PARSERS.get(event_type)
        if parser is None:
            logger.warning(
                "Unknown thread event type %s; skipping. Payload=%r",
                event_type,
                payload,
            )
            return None
        return parser(payload)
    except TypeError as error:
        raise CodexThreadRunError(f"Failed to parse item: {line}") from error


def _parse_thread_started_event(payload: dict[str, object]) -> ThreadEvent:
    return ThreadStartedEvent(
        thread_id=_require_str(payload.get("thread_id"), "thread.started.thread_id"),
    )


def _parse_turn_started_event(payload: dict[str, object]) -> ThreadEvent:
    _ = payload
    return TurnStartedEvent()


def _parse_turn_completed_event(payload: dict[str, object]) -> ThreadEvent:
    usage_payload = _require_dict(payload.get("usage"), "turn.completed.usage")
    usage = Usage(
        input_tokens=_require_int(
            usage_payload.get("input_tokens"),
            "turn.completed.usage.input_tokens",
        ),
        cached_input_tokens=_require_int(
            usage_payload.get("cached_input_tokens"),
            "turn.completed.usage.cached_input_tokens",
        ),
        output_tokens=_require_int(
            usage_payload.get("output_tokens"),
            "turn.completed.usage.output_tokens",
        ),
    )
    return TurnCompletedEvent(usage=usage)


def _parse_turn_failed_event(payload: dict[str, object]) -> ThreadEvent:
    error_payload = _require_dict(payload.get("error"), "turn.failed.error")
    return TurnFailedEvent(
        error=ThreadError(
            message=_require_str(error_payload.get("message"), "turn.failed.error.message"),
        ),
    )


def _parse_item_started_event(payload: dict[str, object]) -> ThreadEvent | None:
    return _parse_item_event(payload, "item.started")


def _parse_item_updated_event(payload: dict[str, object]) -> ThreadEvent | None:
    return _parse_item_event(payload, "item.updated")


def _parse_item_completed_event(payload: dict[str, object]) -> ThreadEvent | None:
    return _parse_item_event(payload, "item.completed")


def _parse_item_event(payload: dict[str, object], event_type: str) -> ThreadEvent | None:
    item = parse_thread_item(payload.get("item"))
    if item is None:
        logger.warning(
            "Unknown thread item for %s; skipping event. Payload=%r",
            event_type,
            payload,
        )
        return None

    if event_type == "item.started":
        return ItemStartedEvent(item=item)
    if event_type == "item.updated":
        return ItemUpdatedEvent(item=item)
    return ItemCompletedEvent(item=item)


def _parse_error_event(payload: dict[str, object]) -> ThreadEvent:
    return ThreadErrorEvent(message=_require_str(payload.get("message"), "error.message"))


_EventParser = Callable[[dict[str, object]], ThreadEvent | None]
_EVENT_PARSERS: dict[str, _EventParser] = {
    "thread.started": _parse_thread_started_event,
    "turn.started": _parse_turn_started_event,
    "turn.completed": _parse_turn_completed_event,
    "turn.failed": _parse_turn_failed_event,
    "item.started": _parse_item_started_event,
    "item.updated": _parse_item_updated_event,
    "item.completed": _parse_item_completed_event,
    "error": _parse_error_event,
}


def parse_thread_item(raw: object) -> ThreadItem | None:
    item = _require_dict(raw, "thread item")
    item_type = _require_str(item.get("type"), "thread item.type")
    parser = _ITEM_PARSERS.get(item_type)
    if parser is None:
        logger.warning("Unknown thread item type %s; skipping. Payload=%r", item_type, item)
        return None
    return parser(item)


def _parse_agent_message_item(item: dict[str, object]) -> ThreadItem:
    return AgentMessageItem(
        id=_require_str(item.get("id"), "agent_message.id"),
        text=_require_str(item.get("text"), "agent_message.text"),
    )


def _parse_reasoning_item(item: dict[str, object]) -> ThreadItem:
    return ReasoningItem(
        id=_require_str(item.get("id"), "reasoning.id"),
        text=_require_str(item.get("text"), "reasoning.text"),
    )


def _parse_command_execution_item(item: dict[str, object]) -> ThreadItem:
    return CommandExecutionItem(
        id=_require_str(item.get("id"), "command_execution.id"),
        command=_require_str(item.get("command"), "command_execution.command"),
        aggregated_output=_require_str(
            item.get("aggregated_output"),
            "command_execution.aggregated_output",
        ),
        status=cast(
            "CommandExecutionStatus",
            _require_str(item.get("status"), "command_execution.status"),
        ),
        exit_code=_optional_int(item.get("exit_code"), "command_execution.exit_code"),
    )


def _parse_file_change_item(item: dict[str, object]) -> ThreadItem:
    raw_changes = _require_list(item.get("changes"), "file_change.changes")
    changes = [
        _parse_file_update_change(raw_change, index) for index, raw_change in enumerate(raw_changes)
    ]
    return FileChangeItem(
        id=_require_str(item.get("id"), "file_change.id"),
        changes=changes,
        status=cast(
            "PatchApplyStatus",
            _require_str(item.get("status"), "file_change.status"),
        ),
    )


def _parse_file_update_change(raw_change: object, index: int) -> FileUpdateChange:
    change_payload = _require_dict(raw_change, f"file_change.changes[{index}]")
    return FileUpdateChange(
        path=_require_str(change_payload.get("path"), f"file_change.changes[{index}].path"),
        kind=cast(
            "PatchChangeKind",
            _require_str(change_payload.get("kind"), f"file_change.changes[{index}].kind"),
        ),
    )


def _parse_mcp_tool_call_item(item: dict[str, object]) -> ThreadItem:
    return McpToolCallItem(
        id=_require_str(item.get("id"), "mcp_tool_call.id"),
        server=_require_str(item.get("server"), "mcp_tool_call.server"),
        tool=_require_str(item.get("tool"), "mcp_tool_call.tool"),
        arguments=item.get("arguments"),
        status=cast(
            "McpToolCallStatus",
            _require_str(item.get("status"), "mcp_tool_call.status"),
        ),
        result=_parse_mcp_result(item.get("result")),
        error=_parse_mcp_error(item.get("error")),
    )


def _parse_web_search_item(item: dict[str, object]) -> ThreadItem:
    return WebSearchItem(
        id=_require_str(item.get("id"), "web_search.id"),
        query=_require_str(item.get("query"), "web_search.query"),
    )


def _parse_todo_list_item(item: dict[str, object]) -> ThreadItem:
    raw_items = _require_list(item.get("items"), "todo_list.items")
    todo_items = [
        _parse_todo_item(raw_todo_item, index) for index, raw_todo_item in enumerate(raw_items)
    ]
    return TodoListItem(
        id=_require_str(item.get("id"), "todo_list.id"),
        items=todo_items,
    )


def _parse_todo_item(raw_todo_item: object, index: int) -> TodoItem:
    todo_payload = _require_dict(raw_todo_item, f"todo_list.items[{index}]")
    return TodoItem(
        text=_require_str(todo_payload.get("text"), f"todo_list.items[{index}].text"),
        completed=_require_bool(
            todo_payload.get("completed"),
            f"todo_list.items[{index}].completed",
        ),
    )


def _parse_error_item(item: dict[str, object]) -> ThreadItem:
    return ErrorItem(
        id=_require_str(item.get("id"), "error.id"),
        message=_require_str(item.get("message"), "error.message"),
    )


_ItemParser = Callable[[dict[str, object]], ThreadItem]
_ITEM_PARSERS: dict[str, _ItemParser] = {
    "agent_message": _parse_agent_message_item,
    "reasoning": _parse_reasoning_item,
    "command_execution": _parse_command_execution_item,
    "file_change": _parse_file_change_item,
    "mcp_tool_call": _parse_mcp_tool_call_item,
    "web_search": _parse_web_search_item,
    "todo_list": _parse_todo_list_item,
    "error": _parse_error_item,
}


def _parse_mcp_result(raw_result: object) -> McpToolCallResult | None:
    if raw_result is None:
        return None

    payload = _require_dict(raw_result, "mcp_tool_call.result")
    content = _require_list(payload.get("content"), "mcp_tool_call.result.content")
    return McpToolCallResult(
        content=content,
        structured_content=payload.get("structured_content"),
    )


def _parse_mcp_error(raw_error: object) -> McpToolCallError | None:
    if raw_error is None:
        return None

    payload = _require_dict(raw_error, "mcp_tool_call.error")
    return McpToolCallError(
        message=_require_str(payload.get("message"), "mcp_tool_call.error.message"),
    )


def _require_dict(value: object, field: str) -> dict[str, object]:
    if not isinstance(value, dict):
        raise TypeError(f"{field} must be an object")
    return cast("dict[str, object]", value)


def _require_list(value: object, field: str) -> list[object]:
    if not isinstance(value, list):
        raise TypeError(f"{field} must be an array")
    return value


def _require_str(value: object, field: str) -> str:
    if not isinstance(value, str):
        raise TypeError(f"{field} must be a string")
    return value


def _require_int(value: object, field: str) -> int:
    if isinstance(value, bool) or not isinstance(value, int):
        raise TypeError(f"{field} must be an integer")
    return value


def _optional_int(value: object, field: str) -> int | None:
    if value is None:
        return None
    return _require_int(value, field)


def _require_bool(value: object, field: str) -> bool:
    if not isinstance(value, bool):
        raise TypeError(f"{field} must be a boolean")
    return value


class TurnState(NamedTuple):
    items: list[ThreadItem]
    final_response: str
    usage: Usage | None
    failure_message: str | None


def initial_turn_state() -> TurnState:
    return TurnState(items=[], final_response="", usage=None, failure_message=None)


def reduce_turn_state(state: TurnState, event: ThreadEvent) -> TurnState:
    final_response = state.final_response
    usage = state.usage
    failure_message = state.failure_message

    if isinstance(event, ItemCompletedEvent):
        state.items.append(event.item)
        if isinstance(event.item, AgentMessageItem):
            final_response = event.item.text
    elif isinstance(event, TurnCompletedEvent):
        usage = event.usage
    elif isinstance(event, TurnFailedEvent):
        failure_message = event.error.message

    return TurnState(
        items=state.items,
        final_response=final_response,
        usage=usage,
        failure_message=failure_message,
    )


def build_turn_or_raise(state: TurnState) -> Turn:
    if state.failure_message is not None:
        raise CodexThreadRunError(state.failure_message)

    return Turn(
        items=state.items,
        final_response=state.final_response,
        usage=state.usage,
    )



from __future__ import annotations

import json
import re
from pathlib import Path
from typing import cast

import pytest

from acodex.exceptions import CodexThreadRunError
from acodex.exec import CodexExec
from acodex.thread import Thread, _close_if_possible
from acodex.types.events import ItemCompletedEvent, ThreadStartedEvent
from acodex.types.input import Input, UserInputLocalImage, UserInputText
from acodex.types.items import AgentMessageItem
from acodex.types.turn_options import OutputSchemaInput
from tests.unit.fake_codex_executable import create_fake_codex_executable


def test_thread_run_streamed_yields_events_and_sets_thread_id(tmp_path: Path) -> None:
    thread = _build_thread(
        tmp_path,
        env={"FAKE_CODEX_MODE": "thread_success", "FAKE_THREAD_ID": "thread-123"},
    )

    streamed = thread.run_streamed("hello")
    events = list(streamed.events)

    assert isinstance(events[0], ThreadStartedEvent)
    assert thread.id == "thread-123"


def test_thread_run_returns_completed_turn_with_final_response_and_usage(tmp_path: Path) -> None:
    thread = _build_thread(
        tmp_path,
        env={
            "FAKE_CODEX_MODE": "thread_success",
            "FAKE_RESPONSES_JSON": json.dumps(["draft", "final answer"]),
        },
    )

    turn = thread.run("hello")

    assert len(turn.items) == 2
    assert turn.final_response == "final answer"
    assert turn.usage is not None
    assert turn.usage.input_tokens == 10
    assert turn.usage.cached_input_tokens == 2
    assert turn.usage.output_tokens == 5


def test_thread_run_raises_codex_thread_run_error_on_turn_failed(tmp_path: Path) -> None:
    message = "CLI failure message"
    thread = _build_thread(
        tmp_path,
        env={
            "FAKE_CODEX_MODE": "thread_failed",
            "FAKE_FAILURE_MESSAGE": message,
        },
    )

    with pytest.raises(CodexThreadRunError, match=re.escape(message)):
        thread.run("hello")


def test_thread_input_normalization_joins_text_and_passes_images_in_order(tmp_path: Path) -> None:
    thread = _build_thread(
        tmp_path,
        env={"FAKE_CODEX_MODE": "normalize_capture"},
        thread_id="thread-resume-42",
    )

    turn = thread.run(
        [
            UserInputText(text="first paragraph"),
            UserInputLocalImage(path="/images/a.png"),
            UserInputText(text="second paragraph"),
            UserInputLocalImage(path="/images/b.png"),
        ],
    )

    payload = cast("dict[str, object]", json.loads(turn.final_response))

    assert payload["stdin"] == "first paragraph\n\nsecond paragraph"
    assert payload["images"] == ["/images/a.png", "/images/b.png"]
    assert payload["resume_id"] == "thread-resume-42"
    assert _extract_flag_values(cast("list[str]", payload["argv"]), "--image") == [
        "/images/a.png",
        "/images/b.png",
    ]


def test_thread_output_schema_file_lifecycle_exhausted_stream(tmp_path: Path) -> None:
    schema: OutputSchemaInput = {"type": "object", "properties": {"ok": {"type": "boolean"}}}
    thread = _build_thread(tmp_path, env={"FAKE_CODEX_MODE": "schema_check"})

    turn = thread.run("hello", output_schema=schema)
    payload = cast("dict[str, object]", json.loads(turn.final_response))
    schema_path = Path(cast("str", payload["schema_path"]))

    assert payload["schema_exists"] is True
    assert payload["schema_payload"] == schema
    assert not schema_path.exists()
    assert not schema_path.parent.exists()


def test_thread_output_schema_file_lifecycle_closed_stream(tmp_path: Path) -> None:
    schema: OutputSchemaInput = {"type": "object", "properties": {"name": {"type": "string"}}}
    thread = _build_thread(tmp_path, env={"FAKE_CODEX_MODE": "schema_check"})

    streamed = thread.run_streamed("hello", output_schema=schema)
    events = streamed.events

    _ = next(events)
    payload_event = next(events)
    assert isinstance(payload_event, ItemCompletedEvent)
    assert isinstance(payload_event.item, AgentMessageItem)
    payload = cast("dict[str, object]", json.loads(payload_event.item.text))
    schema_path = Path(cast("str", payload["schema_path"]))

    assert schema_path.exists()
    close_method = getattr(events, "close", None)
    assert close_method is not None
    close_method()

    assert not schema_path.exists()
    assert not schema_path.parent.exists()


def test_thread_run_streamed_cleanup_when_build_exec_args_raises(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    thread = _build_thread(tmp_path, env={"FAKE_CODEX_MODE": "thread_success"})
    schema: OutputSchemaInput = {"type": "object", "properties": {"ok": {"type": "boolean"}}}
    forced_temp_dir = tmp_path / "forced-schema-dir"

    def fake_mkdtemp(*_: object, **__: object) -> str:
        forced_temp_dir.mkdir(parents=True, exist_ok=True)
        return str(forced_temp_dir)

    monkeypatch.setattr("acodex._internal.output_schema_file.tempfile.mkdtemp", fake_mkdtemp)

    events = thread.run_streamed(cast("Input", {"bad": 1}), output_schema=schema).events
    with pytest.raises(TypeError, match=r"input must be str or list\[UserInput\], got dict"):
        next(events)

    assert not forced_temp_dir.exists()


def test_close_if_possible_is_noop_without_close() -> None:
    _close_if_possible(iter(()))


def _build_thread(
    tmp_path: Path,
    *,
    env: dict[str, str],
    thread_id: str | None = None,
) -> Thread:
    executable = create_fake_codex_executable(tmp_path)
    exec_client = CodexExec(executable_path=str(executable), env=env)
    return Thread(exec=exec_client, options={}, thread_options={}, thread_id=thread_id)


def _extract_flag_values(argv: list[str], flag: str) -> list[str]:
    values: list[str] = []
    index = 0
    while index < len(argv):
        if argv[index] == flag and index + 1 < len(argv):
            values.append(argv[index + 1])
            index += 2
            continue
        index += 1
    return values



from __future__ import annotations

from collections.abc import AsyncIterator, Awaitable, Callable, Iterator
from typing import cast

from typing_extensions import Unpack

from acodex._internal.exec import build_exec_args
from acodex._internal.output_schema_file import UNSET, create_output_schema_file
from acodex._internal.thread_core import (
    build_turn_or_raise,
    initial_turn_state,
    parse_thread_event_jsonl,
    reduce_turn_state,
)
from acodex.exec import AsyncCodexExec, CodexExec
from acodex.types.codex_options import CodexOptions
from acodex.types.events import ThreadEvent, ThreadStartedEvent
from acodex.types.input import Input
from acodex.types.thread_options import ThreadOptions
from acodex.types.turn import (
    AsyncRunStreamedResult,
    RunResult,
    RunStreamedResult,
    StreamedTurn,
)
from acodex.types.turn_options import TurnOptions


class Thread:
    """Represent a thread of conversation with the agent.

    One thread can have multiple consecutive turns.
    """

    def __init__(
        self,
        *,
        exec: CodexExec,  # noqa: A002
        options: CodexOptions,
        thread_options: ThreadOptions,
        thread_id: str | None = None,
    ) -> None:
        self._exec = exec
        self._options = options
        self._id = thread_id
        self._thread_options = thread_options

    @property
    def id(self) -> str | None:
        """Return the ID of the thread.

        The ID is populated after the first turn starts.
        """
        return self._id

    def run_streamed(
        self,
        input: Input,  # noqa: A002
        **turn_options: Unpack[TurnOptions],
    ) -> RunStreamedResult:
        """Provide input to the agent and stream turn events as they are produced.

        Set `turn_options["signal"]` via `event.set()` to request cancellation.
        Use `threading.Event` in synchronous flows.

        Returns:
            A streamed turn result with an iterator of parsed events.

        """

        def event_generator() -> Iterator[ThreadEvent]:
            schema_file = create_output_schema_file(turn_options.get("output_schema", UNSET))
            line_stream: Iterator[str] | None = None
            try:
                exec_args = build_exec_args(
                    input=input,
                    options=self._options,
                    thread_options=self._thread_options,
                    thread_id=self._id,
                    turn_options=turn_options,
                    output_schema_path=schema_file.schema_path,
                )

                line_stream = self._exec.run(exec_args)
                for line in line_stream:
                    event = parse_thread_event_jsonl(line)
                    if event is None:
                        continue

                    if isinstance(event, ThreadStartedEvent):
                        self._id = event.thread_id

                    yield event
            finally:
                if line_stream is not None:
                    _close_if_possible(line_stream)
                schema_file.cleanup()

        return StreamedTurn(events=event_generator())

    def run(
        self,
        input: Input,  # noqa: A002
        **turn_options: Unpack[TurnOptions],
    ) -> RunResult:
        """Provide input to the agent and return the completed turn.

        Set `turn_options["signal"]` via `event.set()` to request cancellation.
        Use `threading.Event` in synchronous flows.

        Returns:
            The completed turn with reduced items, final response, and usage.

        """
        events = self.run_streamed(input, **turn_options).events
        state = initial_turn_state()
        try:
            for event in events:
                state = reduce_turn_state(state, event)
                if state.failure_message is not None:
                    break
        finally:
            _close_if_possible(events)

        return build_turn_or_raise(state)


class AsyncThread:
    """Represent a thread of conversation with the agent.

    One thread can have multiple consecutive turns.
    """

    def __init__(
        self,
        *,
        exec: AsyncCodexExec,  # noqa: A002
        options: CodexOptions,
        thread_options: ThreadOptions,
        thread_id: str | None = None,
    ) -> None:
        self._exec = exec
        self._options = options
        self._id = thread_id
        self._thread_options = thread_options

    @property
    def id(self) -> str | None:
        """Return the ID of the thread.

        The ID is populated after the first turn starts.
        """
        return self._id

    async def run_streamed(
        self,
        input: Input,  # noqa: A002
        **turn_options: Unpack[TurnOptions],
    ) -> AsyncRunStreamedResult:
        """Provide input to the agent and stream turn events as they are produced.

        Set `turn_options["signal"]` via `event.set()` to request cancellation.
        Use `asyncio.Event` in asynchronous flows.

        Returns:
            A streamed turn result with an async iterator of parsed events.

        """

        async def event_generator() -> AsyncIterator[ThreadEvent]:
            schema_file = create_output_schema_file(turn_options.get("output_schema", UNSET))
            line_stream: AsyncIterator[str] | None = None
            try:
                exec_args = build_exec_args(
                    input=input,
                    options=self._options,
                    thread_options=self._thread_options,
                    thread_id=self._id,
                    turn_options=turn_options,
                    output_schema_path=schema_file.schema_path,
                )

                line_stream = self._exec.run(exec_args)
                async for line in line_stream:
                    event = parse_thread_event_jsonl(line)
                    if event is None:
                        continue

                    if isinstance(event, ThreadStartedEvent):
                        self._id = event.thread_id

                    yield event
            finally:
                if line_stream is not None:
                    await _aclose_if_possible(line_stream)
                schema_file.cleanup()

        return StreamedTurn(events=event_generator())

    async def run(
        self,
        input: Input,  # noqa: A002
        **turn_options: Unpack[TurnOptions],
    ) -> RunResult:
        """Provide input to the agent and return the completed turn.

        Set `turn_options["signal"]` via `event.set()` to request cancellation.
        Use `asyncio.Event` in asynchronous flows.

        Returns:
            The completed turn with reduced items, final response, and usage.

        """
        events = (await self.run_streamed(input, **turn_options)).events
        state = initial_turn_state()
        try:
            async for event in events:
                state = reduce_turn_state(state, event)
                if state.failure_message is not None:
                    break
        finally:
            await _aclose_if_possible(events)

        return build_turn_or_raise(state)


def _close_if_possible(iterator: object) -> None:
    close_method = getattr(iterator, "close", None)
    if close_method is None:
        return
    cast("Callable[[], None]", close_method)()


async def _aclose_if_possible(iterator: object) -> None:
    close_method = getattr(iterator, "aclose", None)
    if close_method is None:
        return
    await cast("Callable[[], Awaitable[None]]", close_method)()



from __future__ import annotations

import contextlib
import json
import shutil
import tempfile
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from typing import Final, TypeGuard

from acodex.exceptions import CodexOutputSchemaError
from acodex.types.turn_options import OutputSchemaInput


class _Unset:
    __slots__ = ()


UNSET: Final[_Unset] = _Unset()


@dataclass(frozen=True, slots=True)
class OutputSchemaFile:
    schema_path: str | None
    cleanup: Callable[[], None]


def create_output_schema_file(schema: OutputSchemaInput | _Unset = UNSET) -> OutputSchemaFile:
    if schema is UNSET:
        return OutputSchemaFile(schema_path=None, cleanup=_noop)

    raw_schema: object = schema
    if not _is_json_object(raw_schema):
        raise CodexOutputSchemaError("output_schema must be a plain JSON object")

    schema_object: OutputSchemaInput = raw_schema

    temp_dir = tempfile.mkdtemp(prefix="codex-output-schema-")
    temp_dir_path = Path(temp_dir)
    schema_path = temp_dir_path / "schema.json"

    def cleanup() -> None:
        with contextlib.suppress(Exception):
            shutil.rmtree(temp_dir_path)

    try:
        _write_schema_file(schema_path, schema_object)
    except BaseException:
        cleanup()
        raise

    return OutputSchemaFile(schema_path=str(schema_path), cleanup=cleanup)


def _noop() -> None:
    return None


def _is_json_object(value: object) -> TypeGuard[OutputSchemaInput]:
    return isinstance(value, dict)


def _write_schema_file(path: Path, schema: OutputSchemaInput) -> None:
    path.write_text(json.dumps(schema, ensure_ascii=False), encoding="utf-8")



from __future__ import annotations

import json
import math
import re
from decimal import Decimal
from typing import Any, Final, TypeGuard, cast

from acodex.exceptions import CodexConfigError
from acodex.types.codex_options import CodexConfigObject, CodexConfigValue

_CONFIG_BARE_KEY = re.compile(r"^[A-Za-z0-9_-]+$")
_JS_EXPONENTIAL_UPPER_BOUND: Final[float] = 1e21
_JS_EXPONENTIAL_LOWER_BOUND: Final[float] = 1e-6


def serialize_config_overrides(config_overrides: CodexConfigObject) -> list[str]:
    overrides: list[str] = []
    _flatten_config_overrides(config_overrides, "", overrides=overrides)
    return overrides


def to_config_value(value: CodexConfigValue, path: str) -> str:
    raw_value: Any = value

    if isinstance(raw_value, str):
        return json.dumps(raw_value, ensure_ascii=False)
    if isinstance(raw_value, bool):
        return "true" if raw_value else "false"
    if isinstance(raw_value, int):
        return _to_js_number_string(raw_value, path)
    if isinstance(raw_value, float):
        return _to_js_number_string(raw_value, path)
    if isinstance(raw_value, list):
        return _render_list(raw_value, path)
    if isinstance(raw_value, dict):
        return _render_object(raw_value, path)
    if raw_value is None:
        msg = f"Codex config override at {path} cannot be null"
        raise CodexConfigError(msg)

    type_name = type(raw_value).__name__
    msg = f"Unsupported Codex config override value at {path}: {type_name}"
    raise CodexConfigError(msg)


def _render_list(value: list[CodexConfigValue], path: str) -> str:
    rendered = [to_config_value(item, f"{path}[{index}]") for index, item in enumerate(value)]
    return f"[{', '.join(rendered)}]"


def _render_object(value: CodexConfigObject, path: str) -> str:
    parts: list[str] = []
    for key, child in cast("dict[Any, Any]", value).items():
        if not isinstance(key, str) or not key:
            msg = "Codex config override keys must be non-empty strings"
            raise CodexConfigError(msg)
        child_path = f"{path}.{key}"
        parts.append(f"{_format_config_key(key)} = {to_config_value(child, child_path)}")
    return f"{{{', '.join(parts)}}}"


def _flatten_config_overrides(
    value: CodexConfigValue,
    prefix: str,
    *,
    overrides: list[str],
) -> None:
    raw_value: Any = value
    if not _is_plain_object(raw_value):
        if prefix:
            overrides.append(f"{prefix}={to_config_value(raw_value, prefix)}")
            return

        msg = "Codex config overrides must be a plain object"
        raise CodexConfigError(msg)

    entries = list(cast("dict[Any, Any]", raw_value).items())
    if not prefix and not entries:
        return
    if prefix and not entries:
        overrides.append(f"{prefix}={{}}")
        return

    for key, child in entries:
        if not isinstance(key, str) or not key:
            msg = "Codex config override keys must be non-empty strings"
            raise CodexConfigError(msg)

        path = f"{prefix}.{key}" if prefix else key
        _flatten_config_overrides(child, path, overrides=overrides)


def _is_plain_object(value: CodexConfigValue) -> TypeGuard[CodexConfigObject]:
    return isinstance(value, dict)


def _format_config_key(key: str) -> str:
    return key if _CONFIG_BARE_KEY.fullmatch(key) else json.dumps(key, ensure_ascii=False)


def _to_js_number_string(value: float, path: str) -> str:
    try:
        number = float(value)
    except OverflowError as error:
        msg = f"Codex config override at {path} must be a finite number"
        raise CodexConfigError(msg) from error

    if not math.isfinite(number):
        msg = f"Codex config override at {path} must be a finite number"
        raise CodexConfigError(msg)

    if number == 0:
        return "0"

    sign = "-" if number < 0 else ""
    abs_number = abs(number)
    decimal_number = Decimal(repr(abs_number))

    if abs_number >= _JS_EXPONENTIAL_UPPER_BOUND or abs_number < _JS_EXPONENTIAL_LOWER_BOUND:
        mantissa, exponent = format(decimal_number, "e").split("e")
        if "." in mantissa:
            mantissa = mantissa.rstrip("0").rstrip(".")
        normalized_exponent = _normalize_exponent(exponent)
        return f"{sign}{mantissa}e{normalized_exponent}"

    rendered = format(decimal_number, "f")
    if "." in rendered:
        rendered = rendered.rstrip("0").rstrip(".")
    return f"{sign}{rendered}"


def _normalize_exponent(exponent: str) -> str:
    if exponent.startswith("-"):
        sign = "-"
        digits = exponent[1:]
    elif exponent.startswith("+"):
        sign = "+"
        digits = exponent[1:]
    else:
        sign = "+"
        digits = exponent

    normalized_digits = digits.lstrip("0") or "0"
    return f"{sign}{normalized_digits}"



from __future__ import annotations

import json
import shutil
import tempfile
from pathlib import Path
from typing import Any

import pytest

from acodex._internal import output_schema_file
from acodex.exceptions import CodexOutputSchemaError
from acodex.types.turn_options import OutputSchemaInput


def _set_temp_root(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.setenv("TMPDIR", str(tmp_path))
    monkeypatch.setenv("TMP", str(tmp_path))
    monkeypatch.setenv("TEMP", str(tmp_path))
    tempfile.tempdir = None


def test_unset_schema_returns_noop() -> None:
    schema_file = output_schema_file.create_output_schema_file()
    assert schema_file.schema_path is None
    schema_file.cleanup()


@pytest.mark.parametrize("invalid", [None, [], "x", 1])
def test_rejects_non_object_schema(invalid: Any) -> None:
    with pytest.raises(CodexOutputSchemaError, match="output_schema must be a plain JSON object"):
        output_schema_file.create_output_schema_file(invalid)


def test_writes_schema_json_and_cleanup_removes_it(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    _set_temp_root(monkeypatch, tmp_path)

    schema: OutputSchemaInput = {
        "type": "object",
        "properties": {"name": {"type": "string"}},
        "required": ["name"],
    }

    schema_file = output_schema_file.create_output_schema_file(schema)
    assert schema_file.schema_path is not None

    schema_path = Path(schema_file.schema_path)
    assert schema_path.exists()
    assert json.loads(schema_path.read_text(encoding="utf-8")) == schema

    schema_file.cleanup()
    assert not schema_path.exists()
    assert not schema_path.parent.exists()


def test_cleanup_suppresses_removal_errors(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    _set_temp_root(monkeypatch, tmp_path)

    schema: OutputSchemaInput = {"type": "object"}
    schema_file = output_schema_file.create_output_schema_file(schema)

    def boom(_: Path) -> None:
        raise OSError("boom")

    monkeypatch.setattr(shutil, "rmtree", boom)
    schema_file.cleanup()


def test_write_failure_triggers_cleanup(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    schema: OutputSchemaInput = {"type": "object"}

    temp_dir = tmp_path / "codex-output-schema-fixed"

    def fake_mkdtemp(*, prefix: str) -> str:
        assert prefix == "codex-output-schema-"
        temp_dir.mkdir(parents=True, exist_ok=True)
        return str(temp_dir)

    def fake_write_schema_file(_: Path, __: OutputSchemaInput) -> None:
        raise OSError("boom")

    monkeypatch.setattr(tempfile, "mkdtemp", fake_mkdtemp)
    monkeypatch.setattr(output_schema_file, "_write_schema_file", fake_write_schema_file)

    with pytest.raises(OSError, match="boom"):
        output_schema_file.create_output_schema_file(schema)

    assert not temp_dir.exists()
